# ===================================================
# ===================================================
# Relevant functions for empirical application
# ===================================================
# ===================================================




# ===================================================
# Wrapper function for VaR forecasts
# ===================================================

# Inputs: 
# n: length of rolling window
# alpha: confidence level for risk measure forecasts


wrapper <- function(n, alpha){
   
   no.alpha     <- length(alpha)
   
   ###########################################################################################
   # USD/EUR # USD/EUR # USD/EUR # USD/EUR # USD/EUR # USD/EUR # USD/EUR # USD/EUR # USD/EUR #
   ###########################################################################################
   
   # 1. Prepare data
   
   USDEUR.data  <- read.csv("./Data/DEXUSEU.csv")    # series DEXUSEU downloaded from "fred.stlouisfed.org/series/DEXUSEU"
   USDEUR       <- USDEUR.data$DEXUSEU               # USD/EUR exchange rates from 1999 to 2017
   USDEUR       <- as.numeric(as.character(USDEUR))  # convert "factor" to "numeric"
   USDEUR       <- USDEUR[!is.na(USDEUR)]
   Datum        <- USDEUR.data$DATE                  # read in dates and ...
   Datum        <- as.Date(Datum)                    # ... convert to date format
   lr.USDEUR    <- diff(log(USDEUR))                 # Investigate log-returns
   N            <- length(lr.USDEUR)
   
   
   # 2. Forecast CVaR, CES and CExp
   
   USDEUR.CVaR.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   USDEUR.CVaR.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   USDEUR.CVaR.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   USDEUR.CExp.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   USDEUR.CExp.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   USDEUR.CExp.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   
   for(j in n : (N-1) ){
      forecast <-  VaR.forecast(lr.USDEUR[(j-n+1) : j], alpha, v = 10)
      
      USDEUR.CVaR.H[ , (j-n+1)]   <- forecast$H.Hog18[1, ]
      USDEUR.CVaR.GPD[ , (j-n+1)] <- forecast$GPD[1, ]
      USDEUR.CVaR.NP[ , (j-n+1)]  <- forecast$NP[1, ]
      
      USDEUR.CExp.H[ , (j-n+1)]   <- forecast$H.Hog18[3, ]
      USDEUR.CExp.GPD[ , (j-n+1)] <- forecast$GPD[3, ]
      USDEUR.CExp.NP[ , (j-n+1)]  <- forecast$NP[3, ]
   }
   
   
   # 3. Diebold Mariano (1995) Test of equal predictive ability
   
   X <- lr.USDEUR[(n+1):N]
   SR.USD   <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # save score ratios
   DM95.USD <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # Diebold-Mariano (1995) test for score differences
   
   for(a in 1: no.alpha){
      
      # CVaR
      l1.VaR <- loss.VaR(data = X, VaR = USDEUR.CVaR.H[a, ], alpha[a])
      l2.VaR <- loss.VaR(data = X, VaR = USDEUR.CVaR.GPD[a, ],  alpha[a])
      l3.VaR <- loss.VaR(data = X, VaR = USDEUR.CVaR.NP[a, ],  alpha[a])
      
      SR.USD[1, a] <- sum(l1.VaR) / sum(l3.VaR)   # Hill Ratio
      SR.USD[2, a] <- sum(l2.VaR) / sum(l3.VaR)   # POT Ratio
      
      DM95.USD[1, a] <- dm95.test(l1.VaR, l3.VaR, "two.sided") 
      DM95.USD[2, a] <- dm95.test(l2.VaR, l3.VaR, "two.sided")
      
      # CExp
      l1.Exp <- loss.Exp(data = X, Exp = USDEUR.CExp.H[a, ], alpha[a])
      l2.Exp <- loss.Exp(data = X, Exp = USDEUR.CExp.GPD[a, ],  alpha[a])
      l3.Exp <- loss.Exp(data = X, Exp = USDEUR.CExp.NP[a, ],  alpha[a])
      
      SR.USD[1, a+3] <- sum(l1.Exp) / sum(l3.Exp)   # Hill Ratio
      SR.USD[2, a+3] <- sum(l2.Exp) / sum(l3.Exp)   # POT Ratio
      
      DM95.USD[1, a+3] <- dm95.test(l1.Exp, l3.Exp, "two.sided") 
      DM95.USD[2, a+3] <- dm95.test(l2.Exp, l3.Exp, "two.sided")
   }
   
   
   #########################################################################################
   # WTI Spot # WTI Spot # WTI Spot # WTI Spot # WTI Spot # WTI Spot # WTI Spot # WTI Spot #
   #########################################################################################
   
   # 1. Prepare data
   
   WTI.data  <- read.csv("./Data/DCOILWTICO.csv")     # series downloaded from "fred.stlouisfed.org/series/DCOILWTICO"
   WTI       <- WTI.data$DCOILWTICO            # WTI spot prices from 1998 to 2017
   WTI       <- as.numeric(as.character(WTI))  # convert "factor" to "numeric"
   WTI       <- WTI[!is.na(WTI)]
   Datum     <- WTI.data$DATE                  # read in dates and ...
   Datum     <- as.Date(Datum)                 # ... convert to date format
   lr.WTI    <- diff(log(WTI))                 # Investigate log-returns
   N         <- length(lr.WTI)
   
   
   # 2. Forecast CVaR, CES and CExp
   
   WTI.CVaR.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   WTI.CVaR.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   WTI.CVaR.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   WTI.CExp.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   WTI.CExp.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   WTI.CExp.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   
   for(j in n : (N-1) ){
      forecast <-  VaR.forecast(lr.WTI[(j-n+1) : j], alpha, v = 10)
      
      WTI.CVaR.H[ , (j-n+1)]   <- forecast$H.Hog18[1, ]
      WTI.CVaR.GPD[ , (j-n+1)] <- forecast$GPD[1, ]
      WTI.CVaR.NP[ , (j-n+1)]  <- forecast$NP[1, ]
      
      WTI.CExp.H[ , (j-n+1)]   <- forecast$H.Hog18[3, ]
      WTI.CExp.GPD[ , (j-n+1)] <- forecast$GPD[3, ]
      WTI.CExp.NP[ , (j-n+1)]  <- forecast$NP[3, ]
   }
   
   
   
   # 3. Diebold Mariano (1995) Test of equal predictive ability
   
   X <- lr.WTI[(n+1):N]
   SR.WTI   <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # save score ratios
   DM95.WTI <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # Diebold-Mariano (1995) test for score differences
   
   
   for(a in 1: no.alpha){
      
      # CVaR
      l1.VaR <- loss.VaR(data = X, VaR = WTI.CVaR.H[a, ], alpha[a])
      l2.VaR <- loss.VaR(data = X, VaR = WTI.CVaR.GPD[a, ],  alpha[a])
      l3.VaR <- loss.VaR(data = X, VaR = WTI.CVaR.NP[a, ],  alpha[a])
      
      SR.WTI[1, a] <- sum(l1.VaR) / sum(l3.VaR)   # Hill Ratio
      SR.WTI[2, a] <- sum(l2.VaR) / sum(l3.VaR)   # POT Ratio
      
      DM95.WTI[1, a] <- dm95.test(l1.VaR, l3.VaR, "two.sided") 
      DM95.WTI[2, a] <- dm95.test(l2.VaR, l3.VaR, "two.sided")
      
      # CExp
      l1.Exp <- loss.Exp(data = X, Exp = WTI.CExp.H[a, ], alpha[a])
      l2.Exp <- loss.Exp(data = X, Exp = WTI.CExp.GPD[a, ],  alpha[a])
      l3.Exp <- loss.Exp(data = X, Exp = WTI.CExp.NP[a, ],  alpha[a])
      
      SR.WTI[1, a+3] <- sum(l1.Exp) / sum(l3.Exp)   # Hill Ratio
      SR.WTI[2, a+3] <- sum(l2.Exp) / sum(l3.Exp)   # POT Ratio
      
      DM95.WTI[1, a+3] <- dm95.test(l1.Exp, l3.Exp, "two.sided") 
      DM95.WTI[2, a+3] <- dm95.test(l2.Exp, l3.Exp, "two.sided")
   }
   
   
   
   ##########################################################################################
   # SIE # SIE # SIE # SIE # SIE # SIE # SIE # SIE # SIE # SIE # SIE # 
   ##########################################################################################
   
   
   # 1. Prepare data
   
   SIE.data  <- read.csv("./Data/SIE.csv")
   SIE       <- SIE.data$Adj.Close             # SIE prices from 1998 to 2017
   SIE       <- as.numeric(as.character(SIE))  # convert "factor" to "numeric"
   SIE       <- SIE[!is.na(SIE)]
   Datum     <- SIE.data$Date                  # read in dates and ...
   Datum     <- as.Date(Datum)                 # ... convert to date format
   lr.SIE    <- diff(log(SIE))                 # Investigate log-returns
   N         <- length(lr.SIE)

   
   # 2. Forecast CVaR, CES and CExp

   SIE.CVaR.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   SIE.CVaR.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   SIE.CVaR.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   SIE.CExp.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   SIE.CExp.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   SIE.CExp.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   
   for(j in n : (N-1) ){
      forecast <-  VaR.forecast(lr.SIE[(j-n+1) : j], alpha, v = 10)
      
      SIE.CVaR.H[ , (j-n+1)]   <- forecast$H.Hog18[1, ]
      SIE.CVaR.GPD[ , (j-n+1)] <- forecast$GPD[1, ]
      SIE.CVaR.NP[ , (j-n+1)]  <- forecast$NP[1, ]
      
      SIE.CExp.H[ , (j-n+1)]   <- forecast$H.Hog18[3, ]
      SIE.CExp.GPD[ , (j-n+1)] <- forecast$GPD[3, ]
      SIE.CExp.NP[ , (j-n+1)]  <- forecast$NP[3, ]
   }
   
   
   
   # 3. Diebold Mariano (1995) Test of equal predictive ability
   
   X <- lr.SIE[(n+1):N]
   SR.SIE   <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # save score ratios
   DM95.SIE <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # Diebold-Mariano (1995) test for score differences
   
   
   for(a in 1: no.alpha){
      
      # CVaR
      l1.VaR <- loss.VaR(data = X, VaR = SIE.CVaR.H[a, ], alpha[a])
      l2.VaR <- loss.VaR(data = X, VaR = SIE.CVaR.GPD[a, ],  alpha[a])
      l3.VaR <- loss.VaR(data = X, VaR = SIE.CVaR.NP[a, ],  alpha[a])
      
      SR.SIE[1, a] <- sum(l1.VaR) / sum(l3.VaR)   # Hill Ratio
      SR.SIE[2, a] <- sum(l2.VaR) / sum(l3.VaR)   # POT Ratio
      
      DM95.SIE[1, a] <- dm95.test(l1.VaR, l3.VaR, "two.sided") 
      DM95.SIE[2, a] <- dm95.test(l2.VaR, l3.VaR, "two.sided")
      
      # CExp
      l1.Exp <- loss.Exp(data = X, Exp = SIE.CExp.H[a, ], alpha[a])
      l2.Exp <- loss.Exp(data = X, Exp = SIE.CExp.GPD[a, ],  alpha[a])
      l3.Exp <- loss.Exp(data = X, Exp = SIE.CExp.NP[a, ],  alpha[a])
      
      SR.SIE[1, a+3] <- sum(l1.Exp) / sum(l3.Exp)   # Hill Ratio
      SR.SIE[2, a+3] <- sum(l2.Exp) / sum(l3.Exp)   # POT Ratio
      
      DM95.SIE[1, a+3] <- dm95.test(l1.Exp, l3.Exp, "two.sided") 
      DM95.SIE[2, a+3] <- dm95.test(l2.Exp, l3.Exp, "two.sided")
   }
   
   
   ############################################################################################
   # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # VIX # 
   ############################################################################################
   
   
   # 1. Prepare data
   
   VIX.data  <- read.csv("./Data/VIX.csv")
   VIX       <- VIX.data$Adj.Close             # NASDAQ prices from 1998 to 2017
   VIX       <- as.numeric(as.character(VIX))  # convert "factor" to "numeric"
   VIX       <- VIX[!is.na(VIX)]
   Datum     <- VIX.data$Date                  # read in dates and ...
   Datum     <- as.Date(Datum)                 # ... convert to date format
   lr.VIX    <- diff(log(VIX))                 # Investigate log-returns
   N         <- length(lr.VIX)
   
   
   # 2. Forecast CVaR, CES and CExp
   
   VIX.CVaR.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   VIX.CVaR.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   VIX.CVaR.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   VIX.CExp.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   VIX.CExp.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   VIX.CExp.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   
   for(j in n : (N-1) ){
      forecast <-  VaR.forecast(lr.VIX[(j-n+1) : j], alpha, v = 10)
      
      VIX.CVaR.H[ , (j-n+1)]   <- forecast$H.Hog18[1, ]
      VIX.CVaR.GPD[ , (j-n+1)] <- forecast$GPD[1, ]
      VIX.CVaR.NP[ , (j-n+1)]  <- forecast$NP[1, ]
      
      VIX.CExp.H[ , (j-n+1)]   <- forecast$H.Hog18[3, ]
      VIX.CExp.GPD[ , (j-n+1)] <- forecast$GPD[3, ]
      VIX.CExp.NP[ , (j-n+1)]  <- forecast$NP[3, ]
   }
   
   
   
   # 3. Diebold Mariano (1995) Test of equal predictive ability
   
   X <- lr.VIX[(n+1):N]
   SR.VIX   <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # save score ratios
   DM95.VIX <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # Diebold-Mariano (1995) test for score differences
   
   
   for(a in 1: no.alpha){
      
      # CVaR
      l1.VaR <- loss.VaR(data = X, VaR = VIX.CVaR.H[a, ], alpha[a])
      l2.VaR <- loss.VaR(data = X, VaR = VIX.CVaR.GPD[a, ],  alpha[a])
      l3.VaR <- loss.VaR(data = X, VaR = VIX.CVaR.NP[a, ],  alpha[a])
      
      SR.VIX[1, a] <- sum(l1.VaR) / sum(l3.VaR)   # Hill Ratio
      SR.VIX[2, a] <- sum(l2.VaR) / sum(l3.VaR)   # POT Ratio
      
      DM95.VIX[1, a] <- dm95.test(l1.VaR, l3.VaR, "two.sided") 
      DM95.VIX[2, a] <- dm95.test(l2.VaR, l3.VaR, "two.sided")
      
      # CExp
      l1.Exp <- loss.Exp(data = X, Exp = VIX.CExp.H[a, ], alpha[a])
      l2.Exp <- loss.Exp(data = X, Exp = VIX.CExp.GPD[a, ],  alpha[a])
      l3.Exp <- loss.Exp(data = X, Exp = VIX.CExp.NP[a, ],  alpha[a])
      
      SR.VIX[1, a+3] <- sum(l1.Exp) / sum(l3.Exp)   # Hill Ratio
      SR.VIX[2, a+3] <- sum(l2.Exp) / sum(l3.Exp)   # POT Ratio
      
      DM95.VIX[1, a+3] <- dm95.test(l1.Exp, l3.Exp, "two.sided") 
      DM95.VIX[2, a+3] <- dm95.test(l2.Exp, l3.Exp, "two.sided")
   }
   
   
   
   ############################################################################################
   # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # CAC40 # 
   ############################################################################################
   
   
   # 1. Prepare data
   
   CAC40.data  <- read.csv("./Data/CAC40.csv")
   CAC40       <- CAC40.data$Adj.Close             # NASDAQ prices from 1998 to 2017
   CAC40       <- as.numeric(as.character(CAC40))  # convert "factor" to "numeric"
   CAC40       <- CAC40[!is.na(CAC40)]
   Datum       <- CAC40.data$Date                  # read in dates and ...
   Datum       <- as.Date(Datum)                   # ... convert to date format
   lr.CAC40    <- diff(log(CAC40))                 # Investigate log-returns
   N           <- length(lr.CAC40)
   
   
   # 2. Forecast CVaR, CES and CExp
   
   CAC40.CVaR.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   CAC40.CVaR.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   CAC40.CVaR.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   CAC40.CExp.H   <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   CAC40.CExp.GPD <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   CAC40.CExp.NP  <- matrix(NA, nrow = no.alpha, ncol = (N-n)) 
   
   
   for(j in n : (N-1) ){
      forecast <-  VaR.forecast(lr.CAC40[(j-n+1) : j], alpha, v = 10)
      
      CAC40.CVaR.H[ , (j-n+1)]   <- forecast$H.Hog18[1, ]
      CAC40.CVaR.GPD[ , (j-n+1)] <- forecast$GPD[1, ]
      CAC40.CVaR.NP[ , (j-n+1)]  <- forecast$NP[1, ]
      
      CAC40.CExp.H[ , (j-n+1)]   <- forecast$H.Hog18[3, ]
      CAC40.CExp.GPD[ , (j-n+1)] <- forecast$GPD[3, ]
      CAC40.CExp.NP[ , (j-n+1)]  <- forecast$NP[3, ]
   }
   
   
   
   # 3. Diebold Mariano (1995) Test of equal predictive ability
   
   X <- lr.CAC40[(n+1):N]
   SR.CAC   <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # save score ratios
   DM95.CAC <- matrix(NA, nrow=2, ncol=6, dimnames=list(c("Hill", "POT"), c("CVaR 5%", "1%", "0.5%", "CExp 5%", "1%", "0.5%")))  # Diebold-Mariano (1995) test for score differences
   
   
   for(a in 1: no.alpha){
      
      # CVaR
      l1.VaR <- loss.VaR(data = X, VaR = CAC40.CVaR.H[a, ], alpha[a])
      l2.VaR <- loss.VaR(data = X, VaR = CAC40.CVaR.GPD[a, ],  alpha[a])
      l3.VaR <- loss.VaR(data = X, VaR = CAC40.CVaR.NP[a, ],  alpha[a])
      
      SR.CAC[1, a] <- sum(l1.VaR) / sum(l3.VaR)   # Hill Ratio
      SR.CAC[2, a] <- sum(l2.VaR) / sum(l3.VaR)   # POT Ratio
      
      DM95.CAC[1, a] <- dm95.test(l1.VaR, l3.VaR, "two.sided") 
      DM95.CAC[2, a] <- dm95.test(l2.VaR, l3.VaR, "two.sided")
      
      # CExp
      l1.Exp <- loss.Exp(data = X, Exp = CAC40.CExp.H[a, ], alpha[a])
      l2.Exp <- loss.Exp(data = X, Exp = CAC40.CExp.GPD[a, ],  alpha[a])
      l3.Exp <- loss.Exp(data = X, Exp = CAC40.CExp.NP[a, ],  alpha[a])
      
      SR.CAC[1, a+3] <- sum(l1.Exp) / sum(l3.Exp)   # Hill Ratio
      SR.CAC[2, a+3] <- sum(l2.Exp) / sum(l3.Exp)   # POT Ratio
      
      DM95.CAC[1, a+3] <- dm95.test(l1.Exp, l3.Exp, "two.sided") 
      DM95.CAC[2, a+3] <- dm95.test(l2.Exp, l3.Exp, "two.sided")
   }
   
   
   labNames <- c('CVaR','CExp')
   xlab <- bquote(CVaR ~ .(alpha[1]))
   ylab <- bquote(.(labNames[2]) ~ y^2)
   plot(c(1:10), xlab = xlab, ylab = ylab)
   
   
   
   # Collect all results for Table 5
   
   SR   <- matrix(NA, nrow=10, ncol=6, dimnames=list(c("SIE Hill", "POT", "VIX Hill", "POT", "CAC Hill", "POT", "USD Hill", "POT", "WTI Hill", "POT"), 
                                                     c(bquote(CVaR~.(1-alpha[1])), bquote(.(1-alpha[2])), bquote(.(1-alpha[3])), 
                                                       bquote(CExp~.(1-alpha[1])), bquote(.(1-alpha[2])), bquote(.(1-alpha[3])))))  # save score ratios
   DM95 <- matrix(NA, nrow=10, ncol=6, dimnames=list(c("SIE Hill", "POT", "VIX Hill", "POT", "CAC Hill", "POT", "USD Hill", "POT", "WTI Hill", "POT"),
                                                     c(bquote(CVaR~.(1-alpha[1])), bquote(.(1-alpha[2])), bquote(.(1-alpha[3])), 
                                                       bquote(CExp~.(1-alpha[1])), bquote(.(1-alpha[2])), bquote(.(1-alpha[3])))))  # Diebold-Mariano (1995) test for score differences
   
   SR[1:2, ]  <- SR.SIE
   SR[3:4, ]  <- SR.VIX
   SR[5:6, ]  <- SR.CAC
   SR[7:8, ]  <- SR.USD
   SR[9:10, ] <- SR.WTI
   
   DM95[1:2, ]  <- DM95.SIE
   DM95[3:4, ]  <- DM95.VIX
   DM95[5:6, ]  <- DM95.CAC
   DM95[7:8, ]  <- DM95.USD
   DM95[9:10, ] <- DM95.WTI
   
   return(list(SR = SR, DM95 = DM95))
}




# ===================================================
# Function to generate VaR forecasts
# ===================================================

# Inputs: 
# X: time series
# alpha: confidence level for risk measure forecasts
# v: number of initial fitlered residuals to discard before estimation


VaR.forecast <- function(X, alpha, v){  # VaR forecasts using Hoga (2018+)
  n <- length(X) - v
  
  # 1. Fit model to generate conditional risk forecasts
  # 1.1. Estimate GARCH(1,1) model using standard QMLE for log-returns
  outp      <- garchFit(formula = ~garch(1,1), data = X, init.rec = "mci", cond.dist = "QMLE",
                        include.mean = FALSE,  include.delta = FALSE, include.skew = FALSE,
                        include.shape = FALSE, algorithm = "lbfgsb", trace = FALSE)
  param     <- outp@fit$matcoef
  theta.hat <- as.numeric(param[ ,1])


  # 1.2. Compute residuals U.hat
  X         <- X[(v+1) : (v+n)]
  sigma.hat <- outp@sigma.t[(v+1) : (v+n)]  # discard first v volatility estimates, due to initialization effects
  eps.hat   <- X / sigma.hat

  # 1.3 Volatility estimate
  sigma.hat.n1 <- sqrt( theta.hat[1] + theta.hat[2] * X[n]^2 + theta.hat[3] * sigma.hat[n]^2 )
  
  # # 1. Fit model to generate conditional risk forecasts
  # # 1.1. Estimate GJR-GARCH(1,1) model using standard QMLE for log-returns
  # outp      <- garchFit(formula = ~aparch(1,1), data = X, init.rec = "mci", delta = 2, cond.dist = "norm",
  #                       include.mean = FALSE, include.delta = FALSE, algorithm = "lbfgsb", trace = FALSE)
  # param     <- outp@fit$matcoef
  # theta.hat <- as.numeric(param[ ,1])
  # 
  # 
  # # 1.2. Compute residuals U.hat
  # X         <- X[(v+1) : (v+n)]
  # sigma.hat <- outp@sigma.t[(v+1) : (v+n)]  # discard first v volatility estimates, due to initialization effects
  # eps.hat   <- X / sigma.hat
  # 
  # # 1.3 Volatility estimate
  # sigma.hat.n1 <- sqrt( theta.hat[1] + theta.hat[2] * (abs(X[n]) - theta.hat[3] * X[n])^2 + theta.hat[4] * sigma.hat[n]^2 )
  
  
  #####################################
  ######### Different methods #########
  #####################################
  
  # 2. [Hog18+]
  
  # 2.1 Compute optimal k* using [Dea16+, p. 16] (Tail Index Estimation: Quantile Driven Threshold Selection)
  sort.eps <- sort.int(eps.hat, decreasing = TRUE, method = "quick")
  
  k.min  <- floor(log(n)^2)     # minimal value of k
  k.max  <- floor(4*log(n)^2)   # maximal value of k
  
  sup.H <- rep.int(NA, k.max)
  H     <- rep.int(NA, k.max)
  
  for(k in 1 : k.max){
    
    # 2.1.1 For CVaR estimation
    H[k]   <- 1/k * sum(log( sort.eps[1 : (k+1)] / sort.eps[(k+1)] ))     # Hill estimate based on (k+1) largest order statistics
    sup.H[k]   <- max( abs(sort.eps[2 : (k.max+1)] - (k/(1 : k.max))^( H[k] )   * sort.eps[(k+1)]) )
  }
  k.star.H   <- which.min(sup.H[k.min : k.max]) + k.min - 1
  
  # 2.2 Calculate VaR, ES and expectile estimates using k.star
  # 2.2.1 Calculate extreme value indices
  H   <- H[k.star.H]
  
  # 2.2.2 Calculate UNCONDITIONAL Risk measures
  H.EQ   <- sort.eps[k.star.H + 1] * ((n*(1-alpha)) / k.star.H)^(-H)
  H.ES   <- H.EQ   / (1 - H)
  H.Exp   <- H.EQ   * (1/H   - 1)^(-H)
  

  # 2.3 Calculate CONDITIONAL Risk measures
  CVaR.H.Hog18   <- sigma.hat.n1 * (H.EQ  ) 
  CES.H.Hog18    <- sigma.hat.n1 * (H.ES  )
  CExp.H.Hog18   <- sigma.hat.n1 * (H.Exp  )
  
  
  # 3. POT estimation
  # 3.1 Choice of threshold
  min.U  <- min(eps.hat)
  U      <- eps.hat - min.U + 0.0001                      # shift data to positive half-line
  u.WP   <- as.numeric(quantile(U, probs=0.90, type=1))  # recommended by [CES14] and also [QFP01]
  k.star <- floor(0.10 * n)
  
  # 3.2 Calculate VaR, ES and expectiles (see, e.g., McNeil and Frey (2000, Eqns. (10) & (16)))
  POT     <- fpot(U, threshold = u.WP, model = "gpd", std.err = FALSE)$estimate
  sigma   <- as.numeric( POT[1] )
  xi      <- as.numeric( POT[2] )
  
  VaR.GPD <- u.WP + sigma/xi * ( (n*(1-alpha) / k.star)^(-xi) - 1 )
  ES.GPD  <- VaR.GPD / (1-xi) + (sigma - xi * u.WP) / (1-xi)
  ES.GPD  <- ES.GPD  + min.U - 0.0001     # offset shift to positive half-line
  VaR.GPD <- VaR.GPD + min.U - 0.0001     # offset shift to positive half-line
  Exp.GPD <- VaR.GPD   * (1/max(xi, 0) - 1)^(-max(xi, 0)) # use extrapolation formula from paper
  
  
  # 3.3 Calculate CONDITIONAL Risk measures
  CVaR.GPD <- sigma.hat.n1 * VaR.GPD  
  CES.GPD  <- sigma.hat.n1 * ES.GPD
  CExp.GPD <- sigma.hat.n1 * Exp.GPD
  
  
  # 4. NON-PARAMETRIC estimation
  # 4.1 Calculate VaR, ES and expectiles
  no.alpha <- length(alpha)
  VaR.NP <- ES.NP  <- Exp.NP <- numeric(no.alpha)
  
  for(a in 1 : no.alpha){
    VaR.NP[a] <- sort.int(eps.hat, partial = floor(alpha[a] * n))[floor(alpha[a] * n)]
    ES.NP[a]  <- 1/(1-alpha[a]) * mean( eps.hat * 1*(eps.hat >= VaR.NP[a]) )
    Exp.NP[a] <- Expectile.est(eps.hat, alpha[a])
  }
  
  # 4.2 Calculate CONDITIONAL Risk measures
  CVaR.NP <- sigma.hat.n1 * VaR.NP  
  CES.NP  <- sigma.hat.n1 * ES.NP
  CExp.NP <- sigma.hat.n1 * Exp.NP
  
  
  # 5. Return results
  
  return( list( NP = rbind(CVaR.NP, CES.NP, CExp.NP), 
                GPD = rbind(CVaR.GPD, CES.GPD, CExp.GPD), 
                H.Hog18 = rbind(CVaR.H.Hog18, CES.H.Hog18, CExp.H.Hog18) ) )
}



# ===================================================
# Nonparametric Expectile estimator
# ===================================================

# Inputs: 
# X: time series
# alpha: confidence level for Expectile forecasts

Expectile.est <- function(X, alpha){   # LAWS estimator; see Daouia, Girard and Stupfler (2018, p. 8)
  loss.fct       <- function(y){ return( abs( alpha - 1*(y<=0) ) * y^2) }
  expectile.loss <- function(u){ return( sum(loss.fct(X-u)) ) }
  Exp.est        <- optimise(expectile.loss, interval = c(-100, 100), maximum = FALSE)$minimum
  return(Exp.est)
}






# ===================================================
# VaR Score using the quantile score of Taylor (2017, Eqn. (14))
# ===================================================

# Inputs: 
# data: verifying observations
# VaR: forecasts
# alpha: confidence level


VaR.Backtest <- function(data, VaR, alpha){
  S <- sum( (data - VaR) * (alpha - 1*(data <= VaR)) )
  return(S)
}



# ===================================================
# ES Score using the AL log score of Taylor (2019, Eqn. (19))
# ===================================================

# Inputs: 
# data: verifying observations
# VaR: forecasts
# ES: forecasts
# alpha: confidence level


ES.Backtest <- function(data, VaR, ES, alpha){
  S <- sum( -log((alpha-1) / ES) - (data - VaR) * (alpha - 1*(data <= VaR)) / (alpha * ES) )
  return(S)
}



# ===================================================
# Exp Score using AWLS of Gneiting(2011, Eqn. (27))
# ===================================================

# Inputs: 
# data: verifying observations
# Exp: forecasts
# alpha: confidence level


Exp.Backtest <- function(data, Exp, alpha){
  S <- sum( abs(alpha - 1*(data <= Exp)) * (data - Exp)^2 )
  return(S)
}



# ===================================================
# Exp losses using AWLS of Gneiting(2011, Eqn. (27))
# ===================================================

# Inputs: 
# data: verifying observations
# Exp: forecasts
# alpha: confidence level

loss.Exp <- function(data, Exp, alpha){
  S <- abs(alpha - 1*(data <= Exp)) * (data - Exp)^2
  return(S)
}



# ===================================================
# ES losses using the AL log score of Taylor (2019, Eqn. (19))
# ===================================================

# Inputs: 
# data: verifying observations
# VaR: forecasts
# ES: forecasts
# alpha: confidence level

loss.ES <- function(data, VaR, ES, alpha){
  S <- -log((alpha-1) / ES) - (data - VaR) * (alpha - 1*(data <= VaR)) / (alpha * ES)
  return(S)
}




# ===================================================
# VaR losses using the quantile score of Taylor (2017, Eqn. (14))
# ===================================================

# Inputs: 
# data: verifying observations
# VaR: forecasts
# alpha: confidence level

loss.VaR <- function(data, VaR, alpha){
  S <- (data - VaR) * (alpha - 1*(data <= VaR)) 
  return(S)
}




# ===================================================
# Diebold-Mariano (1995) test of equal predicitve ability
# ===================================================

# Inputs: 
# loss1: losses using risk measure forecasts 1
# loss2: losses using risk measure forecasts 2
# alternative: type of test
# h: number of lags to include in long-run variance estimation


dm95.test <- function (loss1, loss2, alternative = c("two.sided", "less", "greater"), h = 1){
  alternative <- match.arg(alternative)
  d <- c(loss1) - c(loss2)
  d.cov <- acf(d, na.action = na.omit, lag.max = h - 1, type = "covariance",
               plot = FALSE)$acf[, , 1]
  d.var <- sum(c(d.cov[1], 2 * d.cov[-1]))/length(d)
  dv <- d.var
  if (dv > 0)
    STATISTIC <- mean(d, na.rm = TRUE)/sqrt(dv)
  else if (h == 1)
    stop("Variance of DM statistic is zero")
  else {
    warning("Variance is negative, using horizon h=1")
    return(dm95.test(loss1, loss2, alternative, h = 1))
  }
  n <- length(d)
  k <- ((n + 1 - 2 * h + (h/n) * (h - 1))/n)^(1/2)
  STATISTIC <- STATISTIC * k
  names(STATISTIC) <- "DM"
  if (alternative == "two.sided")
    PVAL <- 2 * pt(-abs(STATISTIC), df = n - 1)
  else if (alternative == "less")
    PVAL <- pt(STATISTIC, df = n - 1)
  else if (alternative == "greater")
    PVAL <- pt(STATISTIC, df = n - 1, lower.tail = FALSE)
  return(PVAL)
}



